@@ -4,7 +4,6 @@ import requests
 
 __all__ = ['Resource', 'Collection']
 
-ALL_METHODS = 'HEAD, OPTIONS, GET, PUT, POST, DELETE'
 
 
 class SessionFactory(object):
@@ -16,54 +15,62 @@ class SessionFactory(object):
     }
 
     @classmethod
-    def make(cls, user, password):
+    def make(cls, **credentials):
         session = requests.Session()
-        session.auth = (user, password)
+        session.auth = cls.get_auth(**credentials)
         session.headers.update(cls.default_headers)
 
         return session
 
+    @classmethod
+    def get_auth(cls, **credentials):
+        return (
+            credentials.get('user', ''),
+            credentials.get('password', '')
+        )
 
-class Resource(object):
+
+class UsingOptions(object):
+
+    ALL_METHODS = 'HEAD, OPTIONS, GET, PUT, POST, DELETE'
+
+    def __init__(self, *args, **kwargs):
+        super(UsingOptions, self).__init__()
+        self._meta = {
+            'allowed_methods': UsingOptions.ALL_METHODS,
+            'etag': None,
+            'links': {},
+            'fields': None,
+        }
+
+    @property
+    def response(self):
+        return self._response
+
+    @response.setter
+    def response(self, response):
+        self._response = response
+        self._meta['links'] = response.links
+        self._meta['etag'] = response.headers.get('etag', None)
+        if 'Allow' in response.headers:
+            self._meta['allowed_methods'] = response.headers['Allow']
+
+    def load_options(self):
+        if self._session and self.url:
+            self.response = self._session.options(self.url)
+        else:
+            raise ValueError('Cannot load options for this instance')
+
+
+class Resource(UsingOptions):
     url_attribute_name = 'url'
-    _session = None
     session_factory = SessionFactory
 
+    _response = None
+
     def __repr__(self):
         return '<api_toolkit.Resource type="%s">' % self.__class__
 
-    def __init__(self, data, **kwargs):
-        self.resource_data = data
-        self._links = kwargs.get('links', {})
-        self._session = kwargs.get('session', self._session)
-        self._allowed_methods = kwargs.get('allowed_methods', ALL_METHODS)
-
-        self.prepare_collections()
-
-    @classmethod
-    def load(cls, url, **kwargs):
-        session = kwargs.get('session')
-
-        if session is None:
-            user = kwargs.get('user', '')
-            password = kwargs.get('password', '')
-
-            session = cls.session_factory.make(user, password)
-
-        response = session.get(url)
-        response.raise_for_status()
-
-        instance = cls(
-            data=response.json(),
-            links=response.links,
-            session=session,
-        )
-        instance.url = url
-
-        instance._response = response
-
-        return instance
-
     def __setattr__(self, name, value):
         if (hasattr(self, 'resource_data')
             and self.resource_data.has_key(name)
@@ -79,20 +86,46 @@ class Resource(object):
         except AttributeError:
             return object.__getattribute__(self, 'resource_data')[name]
 
+    def __init__(self, **kwargs):
+        super(Resource, self).__init__(**kwargs)
+        self.resource_data = kwargs
+
+    @property
+    def response(self):
+        return self._response
+
+    @classmethod
+    def from_response(cls, response, session):
+        instance = cls(**response.json())
+        instance._session = session
+        instance.response = response
+
+        return instance
+
+    @classmethod
+    def load(cls, url, **kwargs):
+        session = kwargs.pop('session', cls.session_factory.make(**kwargs))
+
+        response = session.get(url)
+        response.raise_for_status()
+
+        return cls.from_response(response, session)
+
+    @response.setter
+    def response(self, response):
+        self._response = response
+        self._meta['links'] = response.links
+        self._meta['etag'] = response.headers.get('etag', None)
+        self._meta['allowed_methods'] = response.headers.get('Allow', self.ALL_METHODS)
+
+        self.prepare_collections()
+
     @property
     def url(self):
         return self.resource_data.get(self.url_attribute_name)
 
-    @url.setter
-    def url(self, value):
-        self.resource_data[self.url_attribute_name] = value
-
-    @property
-    def etag(self):
-        return self.resource_data.get('etag') or self._response.get('etag')
-
     def prepare_collections(self):
-        for item in self._links.values():
+        for item in self._meta['links'].values():
             link_name = item['rel']
             link_url = item['url']
             link_collection = Collection(
@@ -102,15 +135,23 @@ class Resource(object):
             setattr(self, link_name, link_collection)
 
     def save(self):
-        if 'PUT' not in self._allowed_methods:
+        if 'PUT' not in self._meta['allowed_methods']:
             raise ValueError('This resource cannot be saved.')
 
-        dumped_data = json.dumps(self.resource_data)
+        if self._meta['fields'] is None:
+            dumped_data = json.dumps(self.resource_data)
+        else:
+            filtered_data = dict((k, v)
+                for k, v in self.resource_data.items()
+                if k in self._meta['fields']
+            )
+            dumped_data = json.dumps(filtered_data)
+
         headers = {
             'Content-Length': str(len(dumped_data))
         }
-        if self.resource_data.has_key('etag'):
-            headers.update({'If-Match': self.etag})
+        if self._meta.get('etag'):
+            headers.update({'If-Match': self._meta['etag']})
 
         response = self._session.put(
             self.url,
@@ -119,60 +160,51 @@ class Resource(object):
         )
         response.raise_for_status()
 
-        response = self._session.get(self.url)
-        response.raise_for_status()
+        try:
+            instance = self.__class__.from_response(
+                response=response, session=self._session
+            )
+        except ValueError:
+            instance = self.__class__.load(self.url, session=self._session)
 
-        self.resource_data = response.json()
-        self._links=response.links
-
-        return self
+        return instance
 
     def delete(self):
-        if 'DELETE' not in self._allowed_methods:
+        if 'DELETE' not in self._meta['allowed_methods']:
             raise ValueError('This resource cannot be deleted.')
 
         headers = {}
-        if self.etag:
-            headers.update({'If-Match': self.etag})
+        if self._meta.get('etag'):
+            headers.update({'If-Match': self._meta['etag']})
+
         response = self._session.delete(self.url, headers=headers)
         response.raise_for_status()
 
 
-class Collection(object):
+class Collection(UsingOptions):
     session_factory = SessionFactory
 
     def __repr__(self):
         return '<api_toolkit.Collection type="%s">' % self.__class__
 
     def __init__(self, url, **kwargs):
+        super(Collection, self).__init__(url, **kwargs)
         self.url = url
-        self._session = kwargs.get('session')
+        self._session = kwargs.pop('session', self.session_factory.make(**kwargs))
         self.resource_class = kwargs.get('resource_class', Resource)
 
-        if self._session is None:
-            user = kwargs.get('user', '')
-            password = kwargs.get('password', '')
-
-            self._session = self.session_factory.make(user, password)
-
-        self._allowed_methods = self.discover_allowed_methods()
-
-    def discover_allowed_methods(self):
-        response = self._session.options(self.url)
-        return response.headers.get('Allow', ALL_METHODS)
-
-    def all(self):
-        if 'GET' not in self._allowed_methods:
+    def all(self, load_options=False):
+        if 'GET' not in self._meta['allowed_methods']:
             raise ValueError('This collection is not iterable.')
 
         url = self.url
         while True:
             response = self._session.get(url)
             for item in response.json():
-                instance = self.resource_class(
-                    data=item, session=self._session,
-                    allowed_methods=response.headers.get('Allow', None),
-                )
+                instance = self.resource_class(**item)
+                instance._session = self._session
+                if load_options:
+                    instance.load_options()
                 yield instance
 
             if not response.links.has_key('next'):
@@ -181,9 +213,6 @@ class Collection(object):
             url = response.links['next']['url']
 
     def get(self, identifier, append_slash=True):
-        if 'GET' not in self._allowed_methods:
-            raise ValueError('This collection cannot be loaded.')
-
         if append_slash:
             url_template = '{0}{1}/'
         else:
@@ -193,7 +222,7 @@ class Collection(object):
         return self.resource_class.load(url, session=self._session)
 
     def create(self, **kwargs):
-        if 'POST' not in self._allowed_methods:
+        if 'POST' not in self._meta['allowed_methods']:
             raise ValueError('No items can be created for this collection.')
 
         resource_data = json.dumps(kwargs)
@@ -206,16 +235,13 @@ class Collection(object):
         response.raise_for_status()
         
         try:
-            instance = self.resource_class(
-                data=response.json(),
-                links=response.links,
-                session=self._session,
+            instance = self.resource_class.from_response(
+                response=response, session=self._session
             )
-            instance.url = self.url
-        
-            instance._response = response
-
         except ValueError:
-            instance = self.resource_class.load(response.headers['Location'], session=self._session)
+            instance = self.resource_class.load(
+                url=response.headers['Location'],
+                session=self._session
+            )
 
         return instance
